import scipy as sp
import numpy as np
from copy import deepcopy
import matplotlib.pyplot as plt
import warnings

from lhcsmapi.Time import Time
from lhcsmapi.analysis.diode.DiodeLeadResistanceAnalysis import DiodeLeadResistanceAnalysis
from lhcsmapi.analysis.diode.DiodeLeadResistancePlot import DiodeLeadResistancePlot
from lhcsmapi.dbsignal.ExpSignalAnalysis import ExpSignalAnalysis
from lhcsmapi.dbsignal.SignalProcessing import SignalProcessing
from lhcsmapi.dbsignal.SignalAnalysis import SignalAnalysis
from lhcsmapi.reference.Reference import Reference


def convert_plot_features_map_to_text(features_map):
    max_key_len = max(map(len, features_map))
    return '\n\n'.join(['{}:{} {} {}'.format(key, (max_key_len - len(key)) * ' ', value[0], value[1]) for key, value in
                        features_map.items()])


def calculate_char_time_profile(i_meas_df, timestamp_first_quench=0):
    tau_i_meas_df = deepcopy(i_meas_df)
    signal = tau_i_meas_df.columns[0]
    der_name = 'd{}'.format(signal)
    SignalProcessing.calculate_time_derivative_of_signals(tau_i_meas_df, tau_i_meas_df.index, [signal], [der_name])

    # Calculate time constant as a ratio of a signal by its derivative
    char_time_name = '{0}/d{0}'.format(signal)
    tau_i_meas_df[char_time_name] = -tau_i_meas_df[signal] / sp.signal.medfilt(tau_i_meas_df[der_name], 3)
    tau_i_meas_df.loc[tau_i_meas_df.index < timestamp_first_quench, [char_time_name]] = np.nan
    tau_i_meas_df.drop(columns=[der_name], inplace=True)

    return tau_i_meas_df[char_time_name]


# DIODE
def calculate_resistance_nxcals(circuit_name, i_circuit_sync, u_diode_sync):
    U_DIODE = u_diode_sync.columns[0]

    # Resample circuit current
    t_end = 350 if 'RB' in circuit_name else 150
    i_circuit_sync_resampled = SignalAnalysis.resample(i_circuit_sync, ts=0.1, t0=-50, t_end=t_end)
    if len(i_circuit_sync_resampled) > len(u_diode_sync):
        i_circuit_sync_resampled = i_circuit_sync_resampled.head(len(u_diode_sync))

    # Join current and diode voltage into a single dataframe
    u_diode_sync['I_MEAS'] = i_circuit_sync_resampled.values

    # Recover resistive and diode minimum voltage
    DiodeLeadResistanceAnalysis.calculate_minimum_diode_and_current_leads_voltage(u_diode_sync, U_DIODE)

    # Calculate resistance
    u_diode_sync['RES'] = u_diode_sync.apply(lambda row: 1e+6 * row['U_RES'] / row['I_MEAS'], axis=1, raw=False)

    # Set to 0 resistance values for too low current and with too high diode voltage derivative
    DiodeLeadResistanceAnalysis.set_to_zero_resistance_outside_of_range(u_diode_sync, 'I_MEAS')

    # Filter resistance
    DiodeLeadResistanceAnalysis.filter_resistance(u_diode_sync)

    return u_diode_sync

def calculate_char_time_value(series, t_start_decay, t_end_decay):
    # ToDo: possible duplicate of tau_energy - to be fixed with the DSL
    index_start_decay = series.index.get_loc(t_start_decay, method='nearest')

    index_end_decay = series.index.get_loc(t_end_decay, method='nearest')
    df_decay = series.iloc[index_start_decay:index_end_decay] - series.min()
    df_decay_squared = df_decay.pow(2)
    df_decay_tau = ExpSignalAnalysis.calculate_characteristic_time_in_seconds_with_integral(
        df_decay_squared, exponent=2)
    return df_decay_tau[0]


def plot_earth_current(i_a_df, i_earth_df, i_earth_ref_df, *, title='', hwc_test=''):
    ax1 = i_a_df.plot(title=title, figsize=(15, 7.5), color='C0', grid=True, legend=True)
    ax1.title.set_size(20)
    ax1.set_xlabel("time, [s]", fontsize=15)
    ax1.set_ylabel("I_A [A]", color='C0', fontsize=15)
    ax1.legend(loc='lower left')
    ax1.tick_params(labelsize=15)

    ax2 = ax1.twinx()
    i_earth_df.plot(ax=ax2, xlim=(-1, 5), color='C2', legend=True)
    i_earth_ref_df.plot(ax=ax2, xlim=(-1, 5), color='C3', style='--')
    ax2.set_ylabel("IEARTH [A]", color='C2', fontsize=15)
    ax2.legend(['IEARTH', 'IEARTH ({})'.format(hwc_test)], loc='upper right')
    ax2.tick_params(labelsize=15)
    ax1.set_xlim(-0.1, 0.3)
    # correct first axis
    ax1_ylim_min = int(i_a_df.reindex([0.3], method='nearest', tolerance=.1).values[0])
    ax1_ylim_max = int(i_a_df.reindex([-0.1], method='nearest', tolerance=.1).values[0]) * 1.001
    ax1.set_ylim(ax1_ylim_min, ax1_ylim_max)

    plt.show()

class CircuitAnalysis(object):

    def __init__(self, circuit_type):
        self.circuit_type = circuit_type

    # Plot FGC should go to the PC analysis
    def plot_fgc_current(self, circuit_type, i_meas_df, i_meas_ref_df, tau_i_meas_df, tau_i_meas_ref_df, title='',
                         plot_text=''):

        hwc_test = 'PNO.b2' if circuit_type == 'RB' else 'PNO.b3'
        ylim = (50, 110) if circuit_type == 'RB' else (0, 35)

        ax1 = i_meas_df.plot(color='C0', legend=True, grid=True)
        i_meas_ref_df.plot(ax=ax1, color='C1', style='--', grid=True)
        ax1.set_ylabel("I_MEAS [A]", color='C0', fontsize=15)
        ax1.set_xlabel("time, [s]", fontsize=15)
        ax1.legend(['I_MEAS', 'I_MEAS ({})'.format(hwc_test)], loc='lower left')
        ax1.tick_params(labelsize=16)

        ax2 = ax1.twinx()

        tau_i_meas_df.plot(ax=ax2, title=title, figsize=(15, 10), color='C3', legend=True)
        tau_i_meas_ref_df.plot(ax=ax2, color='C2', legend=True, style='--')
        ax2.title.set_size(20)
        ax2.set_ylabel("-I_MEAS/dI_MEAS/dt [s]", color='C3', fontsize=15)
        ax2.legend(['-I_MEAS/dI_MEAS', '-I_MEAS/dI_MEAS ({})'.format(hwc_test)], loc='upper right')

        ax2.set_ylim(ylim)
        ax2.set_xlim(-50, ax2.get_xlim()[1])

        ax2.annotate(plot_text, xytext=(ax2.get_xlim()[1] * 1.2, ax2.get_ylim()[0]), xy=(20, ax2.get_ylim()[0]),
                     fontsize=15, family='monospace')
        ax2.tick_params(labelsize=16)
        plt.show()

    def plot_earth_current_pcnt(self, i_meas_df, i_earth_pcnt_df, i_earth_pcnt_ref_df, *, title='', hwc_test=''):
        ax1 = i_meas_df.plot(title=title, figsize=(15, 7.5), color='C0', grid=True, legend=True)
        ax1.title.set_size(20)
        ax1.set_xlabel("time, [s]", fontsize=15)
        ax1.legend(loc='lower left')
        ax1.set_ylabel("I_MEAS [A]", color='C0', fontsize=15)
        ax1.tick_params(labelsize=15)

        ax2 = ax1.twinx()
        i_earth_pcnt_df.plot(ax=ax2, color='C2', legend=True)
        i_earth_pcnt_ref_df.plot(ax=ax2, color='C3', yerr=3)
        ax2.legend(['I_EARTH_PCNT', 'I_EARTH_PCNT ({})'.format(hwc_test)], loc='upper right')
        ax2.set_ylabel("I_EARTH_PCNT [%]", color='C2', fontsize=15)
        ax2.tick_params(labelsize=15)
        ax2.set_xlim(0.3, 120)
        plt.show()

    def analyze_i_meas_pc(self, circuit_name, timestamp_fgc_rqd, timestamp_fgc_ref_rqd, timestamp_pic_rqd,
                          i_meas_rqd_df, i_meas_ref_rqd_df):
        if any([i_meas_rqd_df.empty, i_meas_ref_rqd_df.empty]):
            warnings.warn('One of the required signals missing, analysis aborted!')
        else:
            tau_i_meas_rqd_df = calculate_char_time_profile(i_meas_rqd_df)
            tau_i_meas_ref_rqd_df = calculate_char_time_profile(i_meas_ref_rqd_df)

            title = '{} {}: I_MEAS(t) and I_MEAS(t)/dI_MEAS(t) (+PNO.b3)\n'.format(
                Time.to_string_short(timestamp_fgc_rqd), circuit_name)

            tau_i_meas_rqd = ExpSignalAnalysis.calculate_char_time(i_meas_rqd_df, t_start_decay=2, t_end_decay=100)
            tau_i_meas_rqd_ref = ExpSignalAnalysis.calculate_char_time(i_meas_ref_rqd_df, t_start_decay=2,
                                                                       t_end_decay=100)
            tau_rq = Reference.get_tau_range(self.circuit_type)
            if (tau_i_meas_rqd < tau_rq[0]) or (tau_i_meas_rqd > tau_rq[1]):
                warnings.warn(
                    'Characteristic time of FGC {} is outside of {} s range!'.format(circuit_name, tau_rq))

            plot_features = {'Circuit name': [circuit_name, ''],
                             'PIC Abort': [Time.to_string_short(timestamp_pic_rqd), ''],
                             'PC Off': [Time.to_string_short(timestamp_fgc_rqd), ''],
                             'PNO.b3': [Time.to_string_short(timestamp_fgc_ref_rqd), ''],
                             'max(I_MEAS)': [round(i_meas_rqd_df.max().values[0]), 'A'],
                             'max(I_MEAS, PNO.b3)': [round(i_meas_ref_rqd_df.max().values[0]), 'A'],
                             'I_MEAS: tau': [round(tau_i_meas_rqd), 's'],
                             'I_MEAS(PNO.b3):tau': [round(tau_i_meas_rqd_ref), 's']}
            plot_text = convert_plot_features_map_to_text(plot_features)

            self.plot_fgc_current(self.circuit_type, i_meas_rqd_df, i_meas_ref_rqd_df, tau_i_meas_rqd_df,
                                  tau_i_meas_ref_rqd_df, title, plot_text)

    def analyze_i_earth_pc(self, circuit_type, circuit_name, timestamp_fgc_rqd, i_a_rqd_df, i_earth_rqd_df,
                           i_earth_rqd_ref_df):
        # TODO check each signal separately to provide a meaningful warning
        if any([i_a_rqd_df.empty, i_earth_rqd_df.empty, i_earth_rqd_ref_df.empty]):
            warnings.warn('One of the required signals missing, analysis aborted!')
        else:
            hwc_test = 'PNO.b2' if circuit_type == 'RB' else 'PNO.b3'
            title = '{} {}: I_A and I_EARTH (+{})'.format(Time.to_string_short(timestamp_fgc_rqd), circuit_name,
                                                          hwc_test)
            plot_earth_current(i_a_rqd_df, i_earth_rqd_df, i_earth_rqd_ref_df, title=title, hwc_test=hwc_test)

    def analyze_i_earth_pcnt_pc(self, circuit_type, circuit_name, timestamp_fgc_rqd, i_meas_rqd_df, i_earth_pcnt_rqd_df,
                                i_earth_pcnt_rqd_ref_df):
        if any([i_meas_rqd_df.empty, i_earth_pcnt_rqd_df.empty, i_earth_pcnt_rqd_ref_df.empty]):
            warnings.warn('One of the required signals missing, analysis aborted!')
        else:
            hwc_test = 'PNO.b2' if circuit_type == 'RB' else 'PNO.b3'
            title = '{} {}: I_MEAS and I_EARTH (+{})'.format(Time.to_string_short(timestamp_fgc_rqd), circuit_name,
                                                             hwc_test)
            self.plot_earth_current_pcnt(i_meas_rqd_df, i_earth_pcnt_rqd_df, i_earth_pcnt_rqd_ref_df, title=title,
                                         hwc_test=hwc_test)

    # NQPS
    def plot_i_meas_nqps_voltage(self, i_meas_df, nqps_signal_dfs, timestamp_fgc_act, circuit_name, *, signal,
                                 xlim=(-5, 150)):
        title = '{} {} I_MEAS(t) and {}(t)'.format(Time.to_string_short(timestamp_fgc_act), circuit_name,
                                                   signal)
        ax1 = i_meas_df.plot(title=title, color='C0', figsize=(15, 7.5), grid=True)
        ax1.title.set_size(20)
        ax1.legend(['I_MEAS'], loc='lower left')
        ax1.set_xlabel("time, [s]", fontsize=15)
        ax1.set_ylabel("I_MEAS, [A]", fontsize=15, color='C0')
        ax1.tick_params(labelsize=15)

        ax2 = ax1.twinx()
        ax2.set_ylabel("{}, [V]".format(signal), fontsize=15)
        ax2.tick_params(labelsize=15)
        for nqps_signal_df in nqps_signal_dfs:
            if not nqps_signal_df.empty:
                nqps_signal_df.plot(ax=ax2, legend=False)

        ax2.set_xlim(xlim)
        plt.show()

    def analyze_u_diode_nqps(self, circuit_name, timestamp_fgc_rqd, i_meas_rqd_df, u_diode_rqd_dfs, signal):
        if i_meas_rqd_df.empty or (len(u_diode_rqd_dfs) == 0):
            warnings.warn('Required signals missing, analysis aborted!')
        else:
            self.plot_i_meas_nqps_voltage(i_meas_rqd_df, u_diode_rqd_dfs, timestamp_fgc_rqd, circuit_name,
                                          signal=signal)
            self.plot_i_meas_nqps_voltage(i_meas_rqd_df, u_diode_rqd_dfs, timestamp_fgc_rqd, circuit_name,
                                          signal=signal, xlim=(-2, 3))

    # DIODE
    def plot_current_voltage_nxcals(self, magnet, circuit_name, u_diode, timestamp, ax=None, *, title='NXCALS'):
        if ax is None:
            ax1 = u_diode.filter(regex='I_MEAS').plot(title=title, figsize=(15, 7.5), legend=True, lw=2, color='C0')
        else:
            ax1 = u_diode.filter(regex='I_MEAS').plot(ax=ax, title=title, legend=True, lw=2, color='C0')

        ax1.title.set_size(20)
        ax1.set_xlabel("time, [s]", fontsize=20)
        ax1.tick_params(axis='both', which='major', labelsize=15)
        ax1.legend(['I_MEAS'], loc='lower left')
        ax1.set_ylabel("I_MEAS [A]", color='C0', fontsize=20)
        ax1.grid()

        ax2 = ax1.twinx()
        u_diode.filter(regex=magnet).plot(ax=ax2, legend=True, lw=2, color='C2')
        ax2.set_ylabel("U_DIODE [V]", color='C2', fontsize=20)
        ax2.tick_params(axis='both', which='major', labelsize=15)

        ax2.set_ylim([0, ax2.get_ylim()[1]])

        if ax is None:
            plt.show()

    def plot_current_resistance_nxcals(self, magnet, circuit_name, u_diode, timestamp, ax=None, *, title='NXCALS'):
        if ax is None:
            ax1 = u_diode.filter(regex='I_MEAS').plot(title=title, figsize=(15, 7.5), legend=None, lw=2, color='C0')
        else:
            ax1 = u_diode.filter(regex='I_MEAS').plot(ax=ax, title=title, legend=None, lw=2, color='C0')

        ax1.title.set_size(20)
        ax1.set_xlabel("time, [s]", fontsize=20)
        ax1.tick_params(axis='both', which='major', labelsize=15)
        ax1.legend(['I_MEAS'], loc='lower left')
        ax1.set_ylabel("I_MEAS [A]", color='C0', fontsize=20)
        ax1.grid()

        ax2 = ax1.twinx()
        u_diode['RES_filtered'].plot(ax=ax2, legend=True, lw=2, color='C2')
        ax2.set_ylabel("R_DIODE_LEADS(t) (Calculated) [uOhm]", color='C2', fontsize=20)
        ax2.tick_params(axis='both', which='major', labelsize=15)
        ax2.legend(['R_DIODE_LEADS(t)'], loc='upper right')

        x_text = 200 if "MB" in magnet else 100
        res_max_cals, t_res_max_cals, i_res_max_cals = DiodeLeadResistanceAnalysis.extract_res_max_time_current_cals(
            u_diode, 'I_MEAS')
        plot_text = ('r_max = {:.1f} uOhm\ni_r_max = {} A\nt_r_max = {} s'.
                     format(res_max_cals, int(i_res_max_cals), int(t_res_max_cals)))
        ax2.annotate(plot_text, xytext=(x_text, res_max_cals / 2), xy=(t_res_max_cals, res_max_cals),
                     arrowprops=dict(facecolor='white', shrink=0.05), fontsize=15)
        if ax is None:
            plt.show()

    def plot_current_voltage_resistance_pm_nxcals(self, is_quench_in_nqps, source_qds, circuit_name, timestamp_qds,
                                                  u_diode_pm, u_diode_nxcals):
        if is_quench_in_nqps and u_diode_pm is not None:
            fig, ax = plt.subplots(2, 2, figsize=(30, 15))
            title = "Circuit: {}, Magnet: {}, Time Stamp: {}, I_MEAS(t), U_DIODE(t), " \
                    "and R_DIODE_LEADS(t)".format(circuit_name, source_qds, Time.to_string_short(timestamp_qds))
            fig.suptitle(title, fontsize=20)
            # # PM
            DiodeLeadResistancePlot.plot_current_voltage_pm(source_qds, u_diode_pm, timestamp_qds, ax=ax[0, 0])
            DiodeLeadResistancePlot.plot_current_resistance_pm(circuit_name, source_qds, u_diode_pm, timestamp_qds,
                                                               ax=ax[1, 0])
            # # NXCALS
            self.plot_current_voltage_nxcals(source_qds, circuit_name, u_diode_nxcals, timestamp_qds,
                                             ax=ax[0, 1])
            self.plot_current_resistance_nxcals(source_qds, circuit_name, u_diode_nxcals, timestamp_qds,
                                                ax=ax[1, 1])
            plt.show()
        else:
            # # NXCALS
            self.plot_current_voltage_nxcals(source_qds, circuit_name, u_diode_nxcals, timestamp_qds)
            self.plot_current_resistance_nxcals(source_qds, circuit_name, u_diode_nxcals, timestamp_qds)

    def analyze_diode_leads(self, circuit_name, source_timestamp_qds_rqd_df, report_analysis,
                            i_meas_u_diode_u_ref_rqd_pm_dfs, i_meas_u_diode_rqd_nxcals_dfs):
        for index, row in source_timestamp_qds_rqd_df.iterrows():
            source_qds, timestamp_qds = row[['source', 'timestamp']]

            # RQD
            # PM
            i_a_rqd_df, u_diode_a_rqd_df, u_diode_b_rqd_df, u_ref_rqd_df = deepcopy(
                i_meas_u_diode_u_ref_rqd_pm_dfs[index])
            res_diode_rqd_pm_df = DiodeLeadResistanceAnalysis.calculate_resistance_pm(i_a_rqd_df, u_diode_a_rqd_df,
                                                                                      u_diode_b_rqd_df, u_ref_rqd_df,
                                                                                      timestamp_qds, time_corr=0.0)

            # NXCALS
            i_meas_diode_rqd_df, u_diode_rqd_df = deepcopy(i_meas_u_diode_rqd_nxcals_dfs[index])
            res_diode_rqd_nxcals_df = calculate_resistance_nxcals(circuit_name, i_meas_diode_rqd_df, u_diode_rqd_df)

            # Plot
            is_quench_in_nqps = True
            self.plot_current_voltage_resistance_pm_nxcals(is_quench_in_nqps, source_qds, circuit_name, timestamp_qds,
                                                           res_diode_rqd_pm_df, res_diode_rqd_nxcals_df)

            # Maximum resistance value
            res_filtered_rqd_pm_max = res_diode_rqd_pm_df['RES'].max()
            res_filtered_rqd_nxcals_max = res_diode_rqd_nxcals_df['RES_filtered'].max()
            res_filtered_rqd_max = max(res_filtered_rqd_pm_max, res_filtered_rqd_nxcals_max)
            col = 'R_LEADS_RQD [uOhm]' if 'RQD' in circuit_name else 'R_LEADS_RQF [uOhm]'
            report_analysis.loc[index, col] = round(res_filtered_rqd_max, 1)