from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder
from lhcsmapi.reference.Reference import Reference
from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.Time import Time


class CircuitQuery(object):

    def __init__(self, circuit_type, circuit_name):
        self.circuit_type = circuit_type
        self.circuit_name = circuit_name

    def find_timestamp_fgc(self, t_start, t_end, verbose=True):
        if verbose:
            print('Finding PM FGC timestamps of {} from {} to {}.'.format(self.circuit_name, t_start, t_end))

        event_source_timestamp_df = QueryBuilder().with_db('PM').with_circuit_type(self.circuit_type).with_duration(
            t_start=t_start, t_end=t_end) \
            .with_metadata(circuit_name=self.circuit_name, system='PC').event_query().df

        timestamp_fgc = event_source_timestamp_df.loc[0, 'timestamp']
        return timestamp_fgc

    def query_pc_pm(self, timestamp_fgc, timestamp_sync, *, signal_names, verbose=True):
        if verbose:
            print('Querying PM at {} for FGC signals of {}, {}.'.format(timestamp_fgc, self.circuit_name, signal_names))

        dfs = QueryBuilder().with_db('PM').with_circuit_type(self.circuit_type).with_timestamp(timestamp_fgc) \
            .with_metadata(circuit_name=self.circuit_name, system='PC', signal=signal_names).signal_query() \
            .synchronize_time(timestamp_sync).convert_index_to_sec().dfs
        return dfs

    def get_timestamp_ref(self, col='fgcPm'):
        pno_b3_rq = Reference.get_power_converter_reference_fpa(self.circuit_type)
        timestamp_ref = pno_b3_rq[(pno_b3_rq['circuit'] == self.circuit_name)][col].values[0]
        timestamp_ref = Time.to_unix_timestamp(timestamp_ref)

        return timestamp_ref

    def find_source_timestamp_ee(self, t_start, duration=[(10, 's'), (1000, 's')], verbose=True):
        if verbose:
            print('Finding PM EE timestamp of {} at {} for {}.'.format(self.circuit_name, t_start, duration))

        return QueryBuilder().with_db('PM').with_circuit_type(self.circuit_type) \
            .with_duration(t_start=int(t_start), duration=duration) \
            .with_metadata(circuit_name=self.circuit_name, system='EE').event_query().df

    def query_ee_u_dump_res_pm(self, timestamp_ee, timestamp_sync, *, signal_names, verbose=True):
        if verbose:
            print(
                'Querying PM at {} for EE voltage signals of {}, {}.'.format(timestamp_ee, self.circuit_name,
                                                                             signal_names))

        return QueryBuilder().with_db('PM').with_circuit_type(self.circuit_type).with_timestamp(timestamp_ee) \
            .with_metadata(circuit_name=self.circuit_name, system='EE', signal=signal_names).signal_query() \
            .remove_initial_offset(timestamp_ee).synchronize_time(timestamp_sync).convert_index_to_sec().dfs

    def query_ee_t_res_pm(self, timestamp_ee, timestamp_sync, *, signal_names, verbose=True):
        if verbose:
            print('Querying PM at {} for EE temperature signals of {}, {}.'.format(timestamp_ee, self.circuit_name,
                                                                                   signal_names))

        return QueryBuilder().with_db('PM').with_circuit_type(self.circuit_type).with_timestamp(timestamp_ee) \
            .with_metadata(circuit_name=self.circuit_name, system='EE', signal=signal_names).signal_query() \
            .remove_values_for_time_less_than(timestamp_ee).synchronize_time(timestamp_sync).convert_index_to_sec().dfs

    def find_source_timestamp_qds(self, t_start, duration=[(10, 's'), (400, 's')], verbose=True):
        if verbose:
            print('Finding PM iQPS timestamps of {} at {} for {}.'.format(self.circuit_name, t_start, duration))

        return QueryBuilder().with_db('PM').with_circuit_type(self.circuit_type) \
            .with_duration(t_start=int(t_start), duration=duration) \
            .with_metadata(circuit_name=self.circuit_name, system='QDS') \
            .event_query().filter_source().drop_duplicate_source().sort_values(by='timestamp').df

    def find_source_timestamp_nqps(self, t_start, duration=[(10, 's'), (400, 's')], verbose=True):
        if verbose:
            print('Finding PM nQPS timestamps of {} at {} for {}.'.format(self.circuit_name, t_start, duration))

        system = 'DIODE_{}'.format(self.circuit_name.split('.')[0])
        return QueryBuilder().with_db('PM').with_circuit_type(self.circuit_type) \
            .with_duration(t_start=int(t_start), duration=duration) \
            .with_metadata(circuit_name=self.circuit_name, system=system) \
            .event_query().filter_source().drop_duplicate_source().sort_values(by='timestamp').df

    def query_iqps_analog_pm(self, source_timestamp_qds_df, signals, verbose=True):
        iqps_analog_dfs = []
        for index, row in source_timestamp_qds_df.iterrows():
            timestamp_qds = row['timestamp']
            source_qds = row['source']

            if verbose:
                print('Querying PM at {} for iQPS analog signals of {}, {}.'.format(timestamp_qds, self.circuit_name,
                                                                                    signals))

            iqps_analog_df = QueryBuilder().with_db('PM').with_circuit_type(self.circuit_type).with_timestamp(
                timestamp_qds) \
                .with_metadata(circuit_name=self.circuit_name, system='QDS', signal=signals,
                               source=source_qds).signal_query() \
                .synchronize_time(timestamp_qds).convert_index_to_sec().filter_median().dfs
            iqps_analog_dfs.append(iqps_analog_df)

        return iqps_analog_dfs

    def query_iqps_digital_pm(self, source_timestamp_qds_df, signals, verbose=True):
        iqps_digital_dfs = []
        for index, row in source_timestamp_qds_df.iterrows():
            timestamp_qds = row['timestamp']
            source_qds = row['source']

            if verbose:
                print(
                    'Querying PM at {} for iQPS analog signals of {}, {}.'.format(timestamp_qds, self.circuit_name,
                                                                                  signals))

            iqps_digital_df = QueryBuilder().with_db('PM').with_circuit_type(self.circuit_type).with_timestamp(
                timestamp_qds) \
                .with_metadata(circuit_name=self.circuit_name, system='QDS', signal=signals,
                               source=source_qds).signal_query() \
                .synchronize_time(timestamp_qds).convert_index_to_sec().dfs
            iqps_digital_dfs.append(iqps_digital_df)

        return iqps_digital_dfs

    def query_nqps_voltage_pm(self, source_timestamp_qds_df):
        u_nqps_dfs = []

        for index, row in source_timestamp_qds_df.iterrows():
            timestamp_qds = row['timestamp']
            source_qds = row['source']
            qps_crate = MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, source_qds)
            system = 'DIODE_{}'.format(self.circuit_name.split('.')[0])
            signal = 'U_DIODE_{}'.format(self.circuit_name.split('.')[0])
            wildcard = 'MB.{}'.format(source_qds) if self.circuit_type == 'RB' else 'MQ.{}'.format(source_qds)

            source_timestamp_nqps_df = self.find_source_timestamp_nqps_pm(source_qds, timestamp_qds)
            timestamp_nqps = source_timestamp_nqps_df.loc[0, 'timestamp']

            u_diode_rqd_df = QueryBuilder().with_db('PM').with_circuit_type(self.circuit_type).with_timestamp(
                timestamp_nqps).with_metadata(circuit_name=self.circuit_name, system=system, signal=signal,
                                              source=qps_crate,
                                              wildcard={'MAGNET': wildcard}).signal_query().synchronize_time(
                timestamp_nqps).convert_index_to_sec().dfs[0]

            u_ref_rqd_df = QueryBuilder().with_db('PM').with_circuit_type(self.circuit_type).with_timestamp(
                timestamp_nqps).with_metadata(circuit_name=self.circuit_name, system=system, signal='U_REF_N1',
                                              source=qps_crate).signal_query().synchronize_time(
                timestamp_nqps).convert_index_to_sec().dfs[0]
            u_nqps_dfs.append([u_diode_rqd_df, u_ref_rqd_df])

        return u_nqps_dfs

    # DIODE LEADS
    def find_source_timestamp_nqps_pm(self, source_qds, timestamp_qds):
        qps_crate = MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, source_qds)
        system = 'DIODE_{}'.format(self.circuit_name.split('.')[0])

        return QueryBuilder().with_db('PM').with_circuit_type(self.circuit_type).with_duration(t_start=timestamp_qds,
                                                                                               duration=[(600, 's'),
                                                                                                         (600, 's')]) \
            .with_metadata(circuit_name=self.circuit_name, system=system, source=qps_crate).event_query().df

    def query_current_voltage_diode_pm(self, circuit_name, source_qds, timestamp_fgc, timestamp_qds_a,
                                       timestamp_qds_b):

        i_a_rqd_df = QueryBuilder().with_db('PM').with_circuit_type(self.circuit_type).with_timestamp(timestamp_fgc) \
            .with_metadata(circuit_name=circuit_name, system='PC', signal='I_A').signal_query().dfs[0]

        qps_crate = MappingMetadata.get_crate_name_from_magnet_name(self.circuit_type, source_qds)
        system = 'DIODE_{}'.format(circuit_name.split('.')[0])
        signal = 'U_DIODE_{}'.format(circuit_name.split('.')[0])
        wildcard = 'MB.{}'.format(source_qds) if self.circuit_type == 'RB' else 'MQ.{}'.format(source_qds)

        u_diode_a_rqd_df = QueryBuilder().with_db('PM').with_circuit_type(self.circuit_type).with_timestamp(
            timestamp_qds_a).with_metadata(circuit_name=circuit_name, system=system, signal=signal, source=qps_crate,
                                           wildcard={'MAGNET': wildcard}).signal_query().dfs[0]

        u_diode_b_rqd_df = QueryBuilder().with_db('PM').with_circuit_type(self.circuit_type).with_timestamp(
            timestamp_qds_b).with_metadata(circuit_name=circuit_name, system=system, signal=signal, source=qps_crate,
                                           wildcard={'MAGNET': wildcard}).signal_query().dfs[0]

        u_ref_rqd_df = QueryBuilder().with_db('PM').with_circuit_type(self.circuit_type).with_timestamp(
            timestamp_qds_a).with_metadata(circuit_name=circuit_name, system=system, signal='U_REF_N1',
                                           source=qps_crate).signal_query().dfs[0]

        return i_a_rqd_df, u_diode_a_rqd_df, u_diode_b_rqd_df, u_ref_rqd_df

    def query_current_voltage_diode_nxcals(self, circuit_name, source_qds, timestamp_qds, *, spark):
        i_meas_diode_df = QueryBuilder().with_db('NXCALS').with_circuit_type(self.circuit_type) \
            .with_duration(t_start=int(timestamp_qds), duration=[(50, 's'), (150, 's')]) \
            .with_metadata(circuit_name=circuit_name, system='PC', signal='I_MEAS') \
            .signal_query(spark) \
            .synchronize_time(timestamp_qds).convert_index_to_sec().dfs[0]

        system = 'DIODE_{}'.format(circuit_name.split('.')[0])
        signal = 'U_DIODE_{}'.format(circuit_name.split('.')[0])
        wildcard = 'MB.{}'.format(source_qds) if self.circuit_type == 'RB' else 'MQ.{}'.format(source_qds)

        u_diode_rqd_df = \
            QueryBuilder().with_db('NXCALS').with_circuit_type(self.circuit_type).with_duration(
                t_start=int(timestamp_qds),
                duration=[(50, 's'), (150, 's')]) \
                .with_metadata(circuit_name=circuit_name, system=system, signal=signal,
                               wildcard={'MAGNET': wildcard}).signal_query(spark) \
                .synchronize_time(timestamp_qds).convert_index_to_sec().dfs[0]

        return i_meas_diode_df, u_diode_rqd_df

    def query_current_voltage_diode_leads_pm(self, timestamp_fgc_rqd, source_timestamp_qds_df):

        i_meas_u_diode_u_ref_pm_dfs = []

        for index, row in source_timestamp_qds_df.iterrows():
            source_qds, timestamp_qds = row[['source', 'timestamp']]

            source_timestamp_nqps_df = self.find_source_timestamp_nqps_pm(source_qds, timestamp_qds)
            i_meas_u_diode_u_ref_pm_df = self.query_current_voltage_diode_pm(self.circuit_name, source_qds,
                                                                             timestamp_fgc_rqd,
                                                                             source_timestamp_nqps_df.loc[
                                                                                 0, 'timestamp'],
                                                                             source_timestamp_nqps_df.loc[
                                                                                 1, 'timestamp'])

            i_meas_u_diode_u_ref_pm_dfs.append(i_meas_u_diode_u_ref_pm_df)

        return i_meas_u_diode_u_ref_pm_dfs

    def query_current_voltage_diode_leads_nxcals(self, source_timestamp_qds_df, *, spark):

        i_meas_u_diode_nxcals_dfs = []
        for index, row in source_timestamp_qds_df.iterrows():
            source_qds, timestamp_qds = row[['source', 'timestamp']]
            i_meas_u_diode_nxcals_df = self.query_current_voltage_diode_nxcals(self.circuit_name, source_qds,
                                                                               timestamp_qds, spark=spark)
            i_meas_u_diode_nxcals_dfs.append(i_meas_u_diode_nxcals_df)

        return i_meas_u_diode_nxcals_dfs

    def query_u_diode_cals(self, timestamp_fgc, *, ldb):
        system = 'DIODE_{}'.format(self.circuit_name.split('.')[0])
        signal = 'U_DIODE_{}'.format(self.circuit_name.split('.')[0])

        return QueryBuilder().with_db('CALS').with_circuit_type(self.circuit_type).with_duration(
            t_start=int(timestamp_fgc), duration=[(50, 's'), (150, 's')]) \
            .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal, wildcard={'MAGNET': '*'}) \
            .signal_query(ldb) \
            .synchronize_time(timestamp_fgc / 1e9).dfs

    def query_u_earth_cals(self, timestamp_fgc, *, ldb):
        # duplicate of the previous one
        system = 'VF_{}'.format(self.circuit_name.split('.')[0])
        signal = 'U_EARTH_{}'.format(self.circuit_name.split('.')[0])

        return QueryBuilder().with_db('CALS').with_circuit_type(self.circuit_type)\
            .with_duration(t_start=int(timestamp_fgc), duration=[(50, 's'), (150, 's')]) \
            .with_metadata(circuit_name=self.circuit_name, system=system, signal=signal, wildcard={'MAGNET': '*'}) \
            .signal_query(ldb) \
            .synchronize_time(timestamp_fgc / 1e9).dfs
