import warnings
import matplotlib.pyplot as plt
from IPython.display import display, HTML
import pandas as pd

from lhcsmapi.Time import Time
from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.analysis.CircuitAnalysis import CircuitAnalysis, convert_plot_features_map_to_text
from lhcsmapi.analysis.qds.iqds.IqdsRqAnalysis import find_circuit_name_with_first_quenched_magnet, choose_u_qs0, \
    find_t_start_t_end_voltage_slope, calculate_voltage_slope
from lhcsmapi.dbsignal.ExpSignalAnalysis import ExpSignalAnalysis
from lhcsmapi.pyedsl.FeatureBuilder import FeatureBuilder
from lhcsmapi.reference.Reference import Reference


def plot_voltage_qh(source_qds, timestamp_qds, u_hds_rq_df, u_hds_rq_ref_df, is_log=False):
    fig, ax = plt.subplots(figsize=(15, 7))
    title = "Magnet: {}, Time Stamp: {}, U_HDS(t)".format(source_qds,
                                                          Time.to_string_short(timestamp_qds + 1000000))
    fig.suptitle(title, fontsize=20)
    u_hds_rq_df[0].plot(ax=ax, grid=True)
    u_hds_rq_df[1].plot(ax=ax, grid=True)
    u_hds_rq_ref_df[0].plot(ax=ax, grid=True, style='--')
    u_hds_rq_ref_df[1].plot(ax=ax, grid=True, style='--')
    ax.tick_params(labelsize=15)
    ax.set_ylabel("U_HDS, [V]", color='C0', fontsize=15)
    ax.set_xlabel("time, [s]", fontsize=15)
    if is_log:
        plt.yscale('log')
    plt.show()


def plot_i_a_u_dump_res(i_a_df, i_ref_df, u_dump_res_df, timestamp_fgc, timestamp_ee,
                        timestamp_pic, *, title='', plot_text=''):
    i_a_max = i_a_df.max().values[0]
    ylim = (i_a_max - 100, i_a_max + 10)

    ax1 = i_a_df.plot(title=title, xlim=(-0.2, 1), ylim=ylim, color='C0', legend=True, figsize=(15, 7.5),
                      grid=True)
    i_ref_df.plot(ax=ax1, color='C1', marker='o', style='-', xlim=(-0.2, 1), legend=True, grid=True)
    ax1.title.set_size(20)
    ax1.set_ylabel("I_A, [A]", color='C0', fontsize=15)
    ax1.set_xlabel("time, [s]", fontsize=15)
    ax1.legend(loc='lower left')
    ax1.tick_params(labelsize=15)

    ax2 = ax1.twinx()
    u_dump_res_df.plot(ax=ax2, color='C2', legend=True)
    ax2.set_ylabel("U_DUMP_RES, [V]", color='C2', fontsize=15)

    ax2.annotate(plot_text, xytext=(1.1, 0), xy=(0.6, 0), fontsize=15, family='monospace')

    ax2.legend(loc='upper right')
    ax2.annotate('', xy=((timestamp_ee - timestamp_fgc) * 1e-9, 0.2 * ax2.get_ylim()[1]),
                 xytext=((timestamp_pic - timestamp_fgc) * 1e-9, 0.2 * ax2.get_ylim()[1]),
                 arrowprops=dict(arrowstyle='<->'))

    ax2.tick_params(labelsize=15)

    ax2.set_xlim(-0.2, 1)
    # correct first axis
    ax1_ylim_min = int(i_a_df.reindex([1], method='nearest', tolerance=.1).values[0])
    ax1_ylim_max = int(i_a_df.reindex([-0.1], method='nearest', tolerance=.1).values[0]) * 1.001
    ax1.set_ylim(ax1_ylim_min, ax1_ylim_max)

    plt.axvline(x=(timestamp_pic - timestamp_fgc) * 1e-9, color='green', linestyle='--', linewidth=1.5);
    plt.show()


class RQCircuitAnalysis(CircuitAnalysis):

    def __init__(self, circuit_type):
        super().__init__(circuit_type)

    def analyse_timestamp_pic(self, timestamp_pic_rqd, timestamp_pic_rqf):
        if timestamp_pic_rqd != timestamp_pic_rqf:
            warnings.warn(
                'PIC RQD ({}) and RQF ({}) timestamps are different!'.format(timestamp_pic_rqd, timestamp_pic_rqf))
        else:
            print('PIC signals of RQD and RQF are the same.')

    # EE
    def plot_single_ee_fgc(self, u_dump_res_df, i_meas_df, *, title='', plot_text=''):

        ax1 = i_meas_df.plot(title=title, color='C0', legend=True, grid=True)
        ax1.title.set_size(20)
        ax1.set_ylabel("I_MEAS, [A]", color='C0', fontsize=15)
        ax1.set_xlabel("time, [s]", fontsize=15)
        ax1.legend(loc='lower left')
        ax1.tick_params(labelsize=15)

        ax2 = ax1.twinx()
        u_dump_res_df.plot(ax=ax2, figsize=(15, 7.5), color='C2', legend=True)
        ax2.set_ylabel("U_DUMP_RES, [V]", color='C2', fontsize=15)

        ax2.annotate(plot_text, xytext=(max(u_dump_res_df.index.max() * 1.2, i_meas_df.index.max()) * 1.1, 0),
                     xy=(0.6, 0), fontsize=15, family='monospace')
        ax2.legend(loc='upper right')
        ax2.tick_params(labelsize=15)
        ax2.set_xlim((-50, ax2.get_xlim()[1]))
        plt.show()

    def analyze_char_time_u_dump_res_ee(self, circuit_name, timestamp_fgc_rqd, u_dump_res_rqd_df, i_meas_rqd_df):
        if u_dump_res_rqd_df.empty:
            raise ValueError('The required signals missing, analysis aborted!')

        # TODO - to feature builder
        tau_u_dump_res = ExpSignalAnalysis.calculate_char_time(u_dump_res_rqd_df, t_start_decay=1, t_end_decay=100)

        tau_rq = Reference.get_tau_range(self.circuit_type)
        if (tau_u_dump_res < tau_rq[0]) or (tau_u_dump_res > tau_rq[1]):
            warnings.warn('Characteristic time of FGC {} is outside of {} s range!'.format(circuit_name, tau_u_dump_res))

        title = '{} {} EE: I_MEAS(t) and U_DUMP_RES(t)'.format(Time.to_string_short(timestamp_fgc_rqd), circuit_name)

        plot_features = {'Circuit name': [circuit_name, ''],
                         'EE: tau': [round(tau_u_dump_res), 's']}
        plot_text = convert_plot_features_map_to_text(plot_features)

        self.plot_single_ee_fgc(u_dump_res_rqd_df, i_meas_rqd_df, title=title, plot_text=plot_text)

    def analyze_delay_time_u_dump_res_ee(self, circuit_name, timestamp_fgc_rqd, timestamp_pic_rqd, timestamp_ee_rqd,
                                         i_a_rqd_df, i_ref_rqd_df, u_dump_res_rqd_df):
        # ToDo if a signal is empty, throw an error!
        if any([i_a_rqd_df.empty, i_ref_rqd_df.empty, u_dump_res_rqd_df.empty]):
            raise ValueError('One of the required signals missing, analysis aborted!')

        title = '{} {} EE: I_A(t) and U_DUMP_RES(t)'.format(Time.to_string_short(timestamp_fgc_rqd), circuit_name)

        t_delay_ee = round((timestamp_ee_rqd - timestamp_pic_rqd) * 1e-6)
        parameters_ee = Reference.get_energy_extraction_reference_parameters(0, 'RQ')
        t_delay_ee_rq = parameters_ee.loc['t_delay_ee', ['min', 'max']].values
        if (t_delay_ee*1e3 < t_delay_ee_rq[0]) or (t_delay_ee*1e3 > t_delay_ee_rq[1]):
            warnings.warn('Characteristic time of FGC {} is outside of {} s range!'.format(circuit_name, t_delay_ee_rq))

        plot_features = {'Circuit name': [circuit_name, ''],
                         'EE: t_delay': [t_delay_ee, 'ms']}
        plot_text = convert_plot_features_map_to_text(plot_features)

        plot_i_a_u_dump_res(i_a_rqd_df, i_ref_rqd_df, u_dump_res_rqd_df, timestamp_fgc_rqd, timestamp_ee_rqd,
                                 timestamp_pic_rqd, title=title, plot_text=plot_text)

    def analyze_qds(self, circuit_names, report_analysis, iqps_analog_dfs, iqps_digital_dfs, u_nqps_rqd_dfs,
                    u_nqps_rqf_dfs):
        for index, row in report_analysis.iterrows():
            source_qds, timestamp_qds = row[['Position', 'Timestamp iQPS']]
            magnet_name = 'MQ.{}'.format(source_qds)

            u_qs0_ext_rq_df, u_qs0_int_rq_df, u_1_ext_rq_df, u_2_ext_rq_df, u_1_int_rq_df, u_2_int_rq_df = \
                iqps_analog_dfs[
                    index]
            st_magnet_ok_ext_rq_df, st_magnet_ok_int_rq_df, st_nqd0_ext_rq_df, st_nqd0_int_rq_df = iqps_digital_dfs[
                index]
            u_diode_rqd_df, u_ref_rqd_df = u_nqps_rqd_dfs[index]
            u_diode_rqf_df, u_ref_rqf_df = u_nqps_rqf_dfs[index]

            # Analysis
            circuit_name = find_circuit_name_with_first_quenched_magnet(circuit_names, u_diode_rqd_df, u_diode_rqf_df)
            aperture = MappingMetadata.get_rq_aperture(circuit_name, magnet_name)
            u_qs0_aperture_df = choose_u_qs0(aperture, u_qs0_int_rq_df, u_qs0_ext_rq_df)
            t_slope_start, t_slope_end = find_t_start_t_end_voltage_slope(u_qs0_aperture_df)
            du_dt = calculate_voltage_slope(u_qs0_aperture_df, t_slope_start, t_slope_end)
            print('The first aperture to quench is {} of magnet {}'.format(aperture, magnet_name))
            print('The voltage increase is {:.3f} V/s'.format(du_dt))
            report_analysis.loc[index, 'Origin'] = '{}/{}'.format(circuit_name.split('.')[0], aperture)
            report_analysis.loc[index, 'du/dt [V/s]'] = round(du_dt, 2)

            fig, ax = plt.subplots(2, 2, figsize=(30, 15))
            title = "Magnet: {}, Aperture: {}/{}, Time Stamp: {}".format(source_qds, circuit_name.split('.')[0],
                                                                         aperture, Time.to_string_short(timestamp_qds))
            fig.suptitle(title, fontsize=20)

            # iQPS analog
            u_qs0_ext_rq_df.plot(ax=ax[0][0])
            u_qs0_int_rq_df.plot(ax=ax[0][0], grid=True)
            ax[0][0].tick_params(labelsize=15)
            ax[0][0].set_xlim((-0.1, 0.1))
            ax[0][0].axhspan(ymin=-0.1, ymax=+0.1, facecolor='xkcd:yellowgreen')
            ax[0][0].axvspan(xmin=t_slope_start, xmax=t_slope_end, facecolor='xkcd:goldenrod')

            # nQPS analog
            u_diode_rqd_df.plot(ax=ax[0][1])
            u_diode_rqf_df.plot(ax=ax[0][1])
            u_diode_rqf_df.plot(ax=ax[0][1])
            u_ref_rqf_df.plot(ax=ax[0][1], grid=True)
            ax[0][1].tick_params(labelsize=15)

            # iQPS digital
            st_magnet_ok_ext_rq_df.astype(int).plot(ax=ax[1][0])
            st_magnet_ok_int_rq_df.astype(int).plot(ax=ax[1][0], grid=True)
            ax[1][0].tick_params(labelsize=15)
            ax[1][0].set_xlim((-0.1, 0.1))
            st_nqd0_ext_rq_df.astype(int).plot(ax=ax[1][0])
            st_nqd0_int_rq_df.astype(int).plot(ax=ax[1][0], grid=True)
            ax[1][0].tick_params(labelsize=15)

            # nQPS analog - zoom
            u_diode_rqd_df.plot(ax=ax[1][1])
            u_ref_rqd_df.plot(ax=ax[1][1])
            u_diode_rqf_df.plot(ax=ax[1][1])
            u_ref_rqf_df.plot(ax=ax[1][1], grid=True)
            ax[1][1].tick_params(labelsize=15)
            ax[1][1].set_xlim((-0.1, 0.1))
            plt.show()

    def analyze_qh(self, circuit_type, report_analysis, u_hds_rq_dfs, u_hds_rq_ref_dfs):
        for index, row in report_analysis.iterrows():
            source_qds, timestamp_qds = row[['Position', 'Timestamp iQPS']]
            timestamp_qds = Time.to_unix_timestamp(timestamp_qds)
            timestamp_ref = Reference.get_quench_heater_reference_discharge(circuit_type, source_qds)

            u_hds_rq_df = u_hds_rq_dfs[index]
            u_hds_rq_ref_df = u_hds_rq_ref_dfs[index]

            row_feature_df = FeatureBuilder().with_signal(u_hds_rq_df[0], u_hds_rq_df[1]) \
                .calculate_features(features=['first', 'last20mean', 'tau_charge'], index=timestamp_qds)

            row_feature_ref_df = FeatureBuilder().with_signal(u_hds_rq_ref_df[0], u_hds_rq_ref_df[1]) \
                .calculate_features(features=['first', 'last20mean', 'tau_charge'], index=timestamp_ref)

            features_qh_concat = pd.concat([row_feature_df, row_feature_ref_df], axis=0)
            features_abs_diff_qh = features_qh_concat.diff().tail(1).abs()

            # DQAMCNMQ_PMHSU, 2, 810, 1020, 0, 10, 0.003, 20, 5, 0
            is_tau_ok = features_abs_diff_qh.filter(regex='tau').values < 0.003
            is_charge_initial_ok = (row_feature_df.filter(regex='first').values > 810) & \
                                   (row_feature_df.filter(regex='first').values < 1020)
            is_charge_final_ok = (row_feature_df.filter(regex='last').values > 0) & \
                                 (row_feature_df.filter(regex='last').values < 10)
            is_discharge_correct = 'OK' if all(is_tau_ok[0]) & all(is_charge_initial_ok[0]) & all(
                is_charge_final_ok[0]) else 'NOT_OK'

            display(HTML(features_qh_concat.to_html()))
            print('The discharge was automatically rated: {}'.format(is_discharge_correct))
            report_analysis.loc[index, 'QH'] = is_discharge_correct

            plot_voltage_qh(source_qds, timestamp_qds, u_hds_rq_df, u_hds_rq_ref_df)
            plot_voltage_qh(source_qds, timestamp_qds, u_hds_rq_df, u_hds_rq_ref_df, is_log=True)

    def create_timestamp_report(self, ts_pic_rqd, ts_pic_rqf, ts_fgc_rqd, ts_fgc_rqf,
                                source_ts_qds_rqd_df, source_ts_nqps_df, source_ts_dfb_rqd_df, source_ts_dfb_rqf_df):
        timestamp_df = pd.DataFrame(columns=['Source', 'Timestamp'])
        timestamp_df['Source'] = ['PIC_RQD', 'PIC_RQF', 'FGC_RQD', 'FGC_RQF', 'iQPS', 'nQPS']
        timestamp_df['Timestamp'] = [ts_pic_rqd, ts_pic_rqf, ts_fgc_rqd, ts_fgc_rqf,
                                     source_ts_qds_rqd_df.loc[0, 'timestamp'],
                                     source_ts_nqps_df.loc[0, 'timestamp']]

        if not source_ts_dfb_rqd_df.empty:
            ts_leads_rqd = source_ts_dfb_rqd_df.loc[0, 'timestamp']
            timestamp_df.append(pd.DataFrame(columns=['Source', 'Timestamp'], data=[['DFB_RQD', ts_leads_rqd]]))

        if not source_ts_dfb_rqf_df.empty:
            ts_leads_rqf = source_ts_dfb_rqf_df.loc[0, 'timestamp']
            timestamp_df.append(pd.DataFrame(columns=['Source', 'Timestamp'], data=[['DFB_RQF', ts_leads_rqf]]))

        timestamp_df['Date and time'] = timestamp_df.apply(lambda col: Time.to_string_short(col['Timestamp']), axis=1)
        timestamp_df.sort_values(by='Timestamp', inplace=True, ignore_index=True)
        return timestamp_df
