import matplotlib.pyplot as plt
from sklearn.linear_model import LinearRegression
import numpy as np
import warnings


class Assertion(object):
    def set_dfs(self, dfs):
        self.dfs = dfs

    def set_time_range(self, t_start, t_end):
        self.t_start = t_start
        self.t_end = t_end


class AssertionBuilder(object):

    def __init__(self):
        self.assertion = Assertion()

    def with_signal(self, dfs):
        self.assertion.set_dfs(dfs)

        return AssertionBuilderSignal(self.assertion)


def plot_with_value_range(dfs, ylabel, value_min, value_max):
    fig, ax = plt.subplots(figsize=(15, 7))

    for df in dfs:
        df.plot(ax=ax, grid=True)

    ax.legend(loc='upper right')
    ax.set_ylabel(ylabel, fontsize=15)
    ax.tick_params(labelsize=15)
    ax.set_xlabel("time, [s]", fontsize=15)

    ax.axhspan(ymin=value_min, ymax=value_max, facecolor='xkcd:yellowgreen', alpha=0.5)

    y_lim_min = 0.9 * value_min if value_min > 0 else 1.1 * value_min
    y_lim_max = 1.1 * value_max if value_max > 0 else 0.9 * value_max

    ax.set_ylim((y_lim_min, y_lim_max))
    plt.show()


class AssertionBuilderSignal(object):

    def __init__(self, assertion):
        self.assertion = assertion

    def with_time_range(self, t_start, t_end):
        self.assertion.set_time_range(t_start, t_end)

        return AssertionBuilderSignalTimeRange(self.assertion)

    def has_min_max_value(self, value_min, value_max, ylabel=''):
        # Assertion
        for df in self.assertion.dfs:
            col = df.columns[0]
            df_above_thres = df[(df[col] < value_min) & (df[col] > value_max)]
            if len(df_above_thres) > 0:
                warnings.warn('{} outside of the [{}, {}] mV threshold'.format(col, value_min, value_max))

        # Plot
        plot_with_value_range(self.assertion.dfs, ylabel, value_min, value_max)

    def compare_to_reference(self, signal_ref_dfs, abs_margin, scaling=1):
        fig, ax = plt.subplots(figsize=(15, 7))
        for signal_df, signal_ref_df in zip(self.assertion.dfs, signal_ref_dfs):
            col = signal_ref_df.columns[0]
            col_ref = '{}_REF'.format(col)
            col_min = '{}_MIN'.format(col)
            col_max = '{}_MAX'.format(col)
            signal_ref_df = signal_ref_df*scaling
            signal_ref_df = signal_ref_df.rename(columns={col: col_ref})
            idx = signal_ref_df.index.union(signal_df.index)
            signal_comparison_df = signal_ref_df.reindex(idx).interpolate('index')
            signal_comparison_df = signal_comparison_df.join(signal_df).dropna()

            signal_comparison_df[col_min] = signal_comparison_df[col_ref] - abs_margin
            signal_comparison_df[col_max] = signal_comparison_df[col_ref] + abs_margin

            # Compare two curves
            signal_comparison_df['Comparison'] = signal_comparison_df.apply(
                lambda row: (row[col] >= row[col_min]) and (row[col] <= row[col_max]), axis=1)
            if any(signal_comparison_df['Comparison'] == False):
                warnings.warn('Signal {0} outside of range [-{1}, {1}]!'.format(col, abs_margin))

            signal_df.plot(ax=ax)
            signal_ref_df.plot(ax=ax, grid=True, style='--')

            ax.fill_between(signal_comparison_df.index, signal_comparison_df[col_min],
                            signal_comparison_df[col_max], color='xkcd:yellowgreen', alpha=0.5)
            ax.tick_params(labelsize=15)

        plt.show()


def plot_with_time_range(dfs, ylabel, t_start, t_end):
    fig, ax = plt.subplots(figsize=(15, 7))

    for df in dfs:
        df.plot(ax=ax, grid=True)

    ax.legend(loc='upper right')
    ax.set_ylabel(ylabel, fontsize=15)
    ax.tick_params(labelsize=15)
    ax.set_xlabel("time, [s]", fontsize=15)

    for t_s, t_e in zip(t_start, t_end):
        ax.axvspan(xmin=t_s, xmax=t_e, facecolor='xkcd:goldenrod')

    plt.show()


class AssertionBuilderSignalTimeRange(object):

    def __init__(self, assertion):
        self.assertion = assertion

    def has_min_max_variation(self, variation_min_max, ylabel=''):
        t_start = self.assertion.t_start
        t_end = self.assertion.t_end

        # Assertion
        for df in self.assertion.dfs:
            col = df.columns[0]
            for t_s, t_e in zip(t_start, t_end):
                slice_df = df[(df.index > t_s) & (df.index < t_e)]
                variation = slice_df.max() - slice_df.min()
                if variation.values[0] > variation_min_max:
                    warnings.warn('The variation of {} ({:.1f} %) exceeds {} % for constant current from {} to {} s'.format(col,
                                                                                                                    variation_min_max,
                                                                                                                    variation.values[
                                                                                                                        0],
                                                                                                                    t_s,
                                                                                                                    t_e))

        # Plot
        plot_with_time_range(self.assertion.dfs, ylabel, t_start, t_end)

    def has_min_max_slope(self, slope_min, slope_max, ylabel='', scaling=3600 * 1000, unit='mV/h'):
        t_start = self.assertion.t_start
        t_end = self.assertion.t_end

        # Assertion
        for df in self.assertion.dfs:
            col = df.columns[0]
            for t_s, t_e in zip(t_start, t_end):
                slice_df = df[(df.index > t_s) & (df.index < t_e)]
                slope = LinearRegression().fit(np.array(slice_df.index).reshape((-1, 1)), slice_df[col]).coef_[
                            0] * scaling
                if (slope < slope_min) and (slope > slope_max):
                    warnings.warn('The drift of {} is {:.3f} {} for constant current from {} to {} s'.format(col, slope, unit,
                                                                                                     t_s, t_e))

        # Plot
        plot_with_time_range(self.assertion.dfs, ylabel, t_start, t_end)
