import unittest
import numpy as np
import pandas as pd

from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.pyedsl.QueryBuilder import QueryBuilder


class TestQueryBuilder(unittest.TestCase):

    def test_metadata_pm_event_query(self):
        # arrange
        t_start = '2018-12-12 00:00:00+01:00'
        t_end = '2018-12-13 00:00:00+01:00'
        circuit_type = 'RQ'
        circuit_name = 'RQD.A12'

        # act
        metadata_act = QueryBuilder().with_db('PM').with_circuit_type(circuit_type).with_duration(
            t_start=t_start, t_end=t_end) \
            .with_metadata(circuit_name=circuit_name, system='PC').query.metadata

        # assert
        metadata_ref = {'system': 'FGC', 'className': '51_self_pmd', 'source': 'RPHE.UA23.RQD.A12',
                        'I_REF': 'STATUS.I_REF', 'I_MEAS': 'STATUS.I_MEAS', 'V_REF': 'STATUS.V_REF',
                        'V_MEAS': 'STATUS.V_MEAS', 'I_EARTH': 'IEARTH.IEARTH', 'I_EARTH_PCNT': 'STATUS.I_EARTH_PCNT',
                        'I_A': 'IAB.I_A', 'I_B': 'IAB.I_B'}
        self.assertEqual(metadata_ref, metadata_act)

    def test_metadata_pm_signal_query(self):
        # arrange
        timestamp = 1544622149620000000
        circuit_type = 'RQ'
        circuit_name = 'RQD.A12'
        signal = 'I_MEAS'

        # act
        metadata_act = QueryBuilder().with_db('PM').with_circuit_type(circuit_type).with_timestamp(timestamp) \
            .with_metadata(circuit_name=circuit_name, system='PC', signal=signal).query\
            .circuit_name_to_system_to_metadata_and_signal

        # assert
        metadata_ref = {'RQD.A12':
                            {'PC':
                                 {'metadata': {'system': 'FGC', 'className': '51_self_pmd',
                                               'source': 'RPHE.UA23.RQD.A12'},
                                  'signal': ['STATUS.I_MEAS']}}}
        self.assertEqual(metadata_ref, metadata_act)

    def test_metadata_nxcals_signal_query_i_meas_single_circuit(self):
        # arrange
        t_start = '2018-12-12 00:00:00+01:00'
        t_end = '2018-12-13 00:00:00+01:00'
        circuit_type = 'RQ'
        circuit_name = 'RQD.A12'
        signal = 'I_MEAS'

        # act
        metadata_act = QueryBuilder().with_db('NXCALS').with_circuit_type(circuit_type)\
            .with_duration(t_start=t_start, t_end=t_end)\
            .with_metadata(circuit_name=circuit_name, system='PC', signal=signal).query\
            .circuit_name_to_system_to_metadata_and_signal

        # assert
        metadata_ref = {'RQD.A12':
                            {'PC':
                                 {'metadata': {'system': 'CMW', 'device': 'RPHE.UA23.RQD.A12',
                                               'property': 'SUB'},
                                  'signal': ['I_MEAS']}}}
        self.assertEqual(metadata_ref, metadata_act)

    def test_metadata_nxcals_signal_query_i_meas_multiple_circuits(self):
        # arrange
        t_start = '2018-12-12 00:00:00+01:00'
        t_end = '2018-12-13 00:00:00+01:00'
        circuit_type = 'RQ'
        # If circuit name is *, take all circuit names for a circuit_type with SignalMetadata.get_circuit_names(circuit_type)
        circuit_name = ['RQD.A12', 'RQD.A23']
        signal = 'I_MEAS'

        # act
        metadata_act = QueryBuilder().with_db('NXCALS').with_circuit_type(circuit_type)\
            .with_duration(t_start=t_start, t_end=t_end)\
            .with_metadata(circuit_name=circuit_name, system='PC', signal=signal).query\
            .circuit_name_to_system_to_metadata_and_signal

        # assert
        metadata_ref = {'RQD.A12':
                            {'PC':
                                 {'metadata': {'system': 'CMW', 'device': 'RPHE.UA23.RQD.A12',
                                               'property': 'SUB'},
                                  'signal': ['I_MEAS']}},
                        'RQD.A23':
                            {'PC':
                                 {'metadata': {'system': 'CMW', 'device': 'RPHE.UA27.RQD.A23',
                                               'property': 'SUB'},
                                  'signal': ['I_MEAS']}}}
        self.assertEqual(metadata_ref, metadata_act)

    def test_metadata_nxcals_signal_query_u_res_multiple_circuits(self):
        # arrange
        t_start = '2018-12-12 00:00:00+01:00'
        t_end = '2018-12-13 00:00:00+01:00'
        circuit_type = 'RB'
        # If circuit name is *, take all circuit names for a circuit_type with SignalMetadata.get_circuit_names(circuit_type)
        circuit_name = 'RB.A12'
        signal = 'U_MAG'
        # if value of the wildcard (here: 'DCBA.B34R3.L') is '*' take all components from MappingMetadata.get_busbars_for_circuit_name(circuit_type, circuit_names)
        wildcard = {'BUSBAR': 'DCBA.B34R3.L'}

        # act
        metadata_act = QueryBuilder().with_db('NXCALS').with_circuit_type(circuit_type) \
            .with_duration(t_start=t_start, t_end=t_end) \
            .with_metadata(circuit_name=circuit_name, system='BUSBAR', signal=signal,
                           wildcard=wildcard).query \
            .circuit_name_to_system_to_metadata_and_signal

        # assert
        metadata_ref =  {'RB.A12':
                             {'BUSBAR':
                                  {'metadata': {'system': 'WINCCOA'},
                                   'signal': ['DCBA.B34R3.L:U_MAG']}}}

        self.assertEqual(metadata_ref, metadata_act)

    def test_feature_query_nxcals(self):
        pass
        """
        QueryBuilder().with_db('NXCALS').with_circuit_type(circuit_type)\
            .with_duration(t_start=t_start, t_end=t_end)\
            .with_metadata(circuit_name=circuit_name, system='PC', signal=signal).feature_query(features=['mean', 'max'])
        """

